import numpy
import math
import time
import scipy.stats


def forward_procedure(A, B, PI, O, scaling = True, discrete=False):
    T = len(O)
    N = len(B)
    alpha = numpy.zeros((N,  T))
    C = numpy.zeros(T)

    if scaling:
        C = numpy.zeros(T)

    alpha[:,0] = PI * [B[i](O[0]) for i in range(N)]

    C[0] = 1.0/numpy.sum(alpha[:,0])
    alpha[:,0] = C[0] * alpha[:,0]

    for t in xrange(1, T):
        b_o = [B[i](O[t]) for i in range(N)]
        alpha[:,t] = numpy.dot(alpha[:,t-1], numpy.exp(A)) * b_o

        C[t] = 1.0/numpy.sum(alpha[:,t])
        
        alpha[:,t] = alpha[:,t] * C[t]
    
    
    log_Prob_Obs = - (numpy.sum(numpy.log(C)))
    return(( log_Prob_Obs, alpha, C ))



def backward_procedure(A, B, PI, O, C=None, discrete=False):

    N = len(B)
    T = len(O)
    beta = numpy.zeros((N,T))


    beta[:,T-1] = 1.0
    if C!=None: beta[:,T-1] = beta[:,T-1] * C[T-1]


    for t in range(T-2, -1, -1):
        b_o = [B[i](O[t]) for i in range(N)]
        beta[:,t] = numpy.dot(numpy.exp(A), (b_o * beta[:,t+1] ) )

        if C!=None:
            beta[:,t] = beta[:,t] * C[t]

    return(beta)




def viterbi(A, B, PI, O, scaling=True, discrete=False):
    N=len(B)    
    T = len(O)
    delta = numpy.zeros((N, T))

    b_o = [B[i](O[0]) for i in range(N)]
    if scaling:
        delta[:,0] = numpy.log(PI + numpy.finfo(float).eps) + numpy.log(b_o + numpy.finfo(float).eps)

    Q = numpy.zeros((N, T))
    

    for t in range(1, T):
        b_o = [B[i](O[t]) for i in range(N)]
        nus = delta[:, t-1] + A
        delta[:,t] = nus.max(1) + numpy.log(b_o + numpy.finfo(float).eps)
        Q[:,t] = nus.argmax(1)

    Q_opt = [numpy.argmax(delta[:,T-1])]
    for t in range(T-2, -1, -1):
        Q_opt.insert(0, int(Q[Q_opt[0],t+1]))

    return((Q_opt, delta, Q))



from scipy.stats import entropy
def UncertaintyMeasure(pp):
    values= numpy.exp(list(pp.values()))
    return entropy(values)
    
    max_value = max(numpy.exp(list(pp.values())))
    if max_value < 1:
        return 1
    else:
        return 0
    
import operator


def Peek_Viterbi(obs, states, log_start_p, log_trans_p, B, label_inverse_map, adjustment, gamma, L):
        
    V = [{}]
    for st in states:
        V[0][st] = {"log_prob": log_start_p[st] + adjustment + numpy.log(B[label_inverse_map[st]](obs[0]) + numpy.finfo(float).eps), "prev": None}
    for t in range(1, min(L+1, len(obs))):
        V.append({})
        for st in states:
            
            max_log_prob = max(V[t-1][prev_st]["log_prob"]+ gamma*log_trans_p[prev_st][st] + \
                               gamma*(adjustment + numpy.log(B[label_inverse_map[st]](obs[t]) + numpy.finfo(float).eps)) for prev_st in states)
            
            for prev_st in states:
                if (V[t-1][prev_st]["log_prob"] + gamma*log_trans_p[prev_st][st]
                         + gamma*(adjustment + numpy.log(B[label_inverse_map[st]](obs[t]) + numpy.finfo(float).eps)))  == max_log_prob:
                    V[t][st] = {"log_prob": max_log_prob, "prev": prev_st}
                    break
    opt = []
    
    # The highest log probability
    max_log_prob = max(value["log_prob"] for value in V[-1].values())
    previous = None
    # Get most probable state and its backtrack
    for st, data in V[-1].items():
        if data["log_prob"] == max_log_prob:
            opt.append(st)
            previous = st
            break
    # Follow the backpath till the first observation
    for t in range(len(V) - 2, -1, -1):
        opt.insert(0, V[t + 1][previous]["prev"])
        previous = V[t + 1][previous]["prev"]
        
    return opt[0]
 
    
    
def PeekSearch(B, label_inverse_map, adjustment, actual_states, log_transition_probs, log_start_p, obs, L):
    
    gamma = numpy.power(1/(L+1), (1/L))
    
    log_start = log_start_p.copy()
    predicted_tags = []
    for j in range(len(obs)):
        next_state = Peek_Viterbi(obs[j:j+L+1], actual_states, log_start, log_transition_probs, \
                                  B, label_inverse_map, adjustment, gamma, L)
        predicted_tags.append(next_state)
        log_start = {st:log_transition_probs[next_state][st] for st in actual_states}

    return predicted_tags


def PeekSearch_Gamma1(B, label_inverse_map, adjustment, actual_states, log_transition_probs, log_start_p, obs, L):
    
    gamma =  1.0
    
    log_start = log_start_p.copy()
    predicted_tags = []
    for j in range(len(obs)):
        next_state = Peek_Viterbi(obs[j:j+L+1], actual_states, log_start, log_transition_probs, \
                                  B, label_inverse_map, adjustment, gamma, L)
        predicted_tags.append(next_state)
        log_start = {st:log_transition_probs[next_state][st] for st in actual_states}

    return predicted_tags


def hash_genes(path):
    hash = {}
    for line in open(path):
        if line.startswith("#"): continue
        tmp = line.strip().split("\t")
        if tmp[2] != "gene": continue
        features = dict([tuple(f.split("=")) for f in tmp[8].split(";")])
        start, end = int(tmp[3]), int(tmp[4])
        for i in range(start,end+1):
            if i not in hash:
                hash[i] = features.get("ID", "missing")
    return hash

def get_name(path):
    rv2name = {}
    for line in open(path):
        if line.startswith("#"): continue
        tmp = line.split("\t")
        if tmp[2] != "gene": continue
        features = dict([tuple(f.split("=")) for f in tmp[8].strip().split(";")])
        rv2name[features.get("ID", "missing")] = features.get("Name", "-")
    return(rv2name)



def calculate_pins(reads):

    non_ess_reads = []
    temp = []
    for rd in reads:

        if rd >=1:
            if len(temp) < 10: non_ess_reads.extend(temp)
            non_ess_reads.append(rd)
            temp = []
        else:
            temp.append(rd)

    return(sum([1 for rd in non_ess_reads if rd >= 1])/float(len(non_ess_reads)) )



### Gumbel/EVD Variance and Expected run funnctions
def VarR(n,p):
    """ VarR_n =  (pi^2)/(6*ln(1/p)^2) + 1/12 + r2(n) + E2(n) (Schilling, 1990)"""

    r2 = getR2(n)
    E2 = getE2(n)

    A = math.pow(math.pi,2.0)/(6* math.pow(math.log(1.0/p),2.0))
    V = A + 1/12.0 + r2 + E2

    return ( V )


#### Computes log-probability for a given sequence of predicted state labels. 
def compute_log_probability(actual_states, predicted_tags, obs, log_start_p, log_trans_p, B, label_inverse_map, adjustment = 64):
    
    log_probability = log_start_p[actual_states[predicted_tags[0]]] + numpy.log(B[predicted_tags[0]](obs[0]) \
                                                                               + numpy.finfo(float).eps) - adjustment
    
    for i in range(1, len(predicted_tags)):
        previous_tag = actual_states[predicted_tags[i-1]]
        current_tag = actual_states[predicted_tags[i]]
        log_probability += log_trans_p[previous_tag][current_tag] + \
                            numpy.log(B[label_inverse_map[current_tag]](obs[i]) + numpy.finfo(float).eps) - adjustment

        
    return log_probability


def ExpectedRuns(n,p):
    """ER_n =  log(1/p)(nq) + gamma/ln(1/p) -1/2 + r1(n) + E1(n) (Schilling, 1990)"""

    q = 1-p
    gamma = getGamma()
    r1 = getR1(n)
    E1 = getE1(n)


    A = math.log(n*q,1.0/p)
    B = gamma/math.log(1.0/p)
    ER = A + B -0.5 + r1 + E1

    return( ER )


def getGamma():
    """Euler-Mascheroni constant ~ 0.577215664901 """
    return(0.5772156649015328606)


def getR1(n):
    """Small Correction term. Defaults to 0.000016 for now"""
    return(0.000016)

def getR2(n):
    """Small Correction term. Defaults to 0.00006 for now"""
    return(0.00006)

def getE1(n):
    """Small Correction term. Defaults to 0.01 for now"""
    return(0.01)

def getE2(n):
    """Small Correction term. Defaults to 0.01 for now"""
    return(0.01)


